/*
 * location-module
 *
 * Copyright (c) 2010-2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Youngae Kang <youngae.kang@samsung.com>, Minjune Kim <sena06.kim@samsung.com>
 *          Genie Kim <daejins.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "log.h"
#include "location-osm.h"

static GHashTable*
_get_geoclue_addr_from_location_addr (const LocationAddress *address)
{
	if (!address) return NULL;

	GHashTable *geoclue_addr = geoclue_address_details_new();
	if (address->street) {
		if (address->building_number) {
			char *street_addr = g_strdup_printf ("%s %s", address->building_number, address->street);
			geoclue_address_details_insert (geoclue_addr, GEOCLUE_ADDRESS_KEY_STREET, street_addr);
			g_free (street_addr);
		} else geoclue_address_details_insert (geoclue_addr, GEOCLUE_ADDRESS_KEY_STREET, address->street);
	}
	if (address->district)     geoclue_address_details_insert (geoclue_addr, GEOCLUE_ADDRESS_KEY_AREA, address->district);
	if (address->city)         geoclue_address_details_insert (geoclue_addr, GEOCLUE_ADDRESS_KEY_LOCALITY, address->city);
	if (address->state)        geoclue_address_details_insert (geoclue_addr, GEOCLUE_ADDRESS_KEY_REGION, address->state);
	if (address->country_code) geoclue_address_details_insert (geoclue_addr, GEOCLUE_ADDRESS_KEY_COUNTRY, address->country_code);
	if (address->postal_code)  geoclue_address_details_insert (geoclue_addr, GEOCLUE_ADDRESS_KEY_POSTALCODE, address->postal_code);

	return geoclue_addr;
}

static LocationAddress*
_get_location_addr_from_geoclue_addr (GHashTable *geoclue_addr)
{
	if (!geoclue_addr) return NULL;

	LocationAddress *addr = NULL;

	gchar *country, *region, *locality, *postalcode, *area, *street;

	country = g_hash_table_lookup (geoclue_addr, GEOCLUE_ADDRESS_KEY_COUNTRY);
	region = g_hash_table_lookup (geoclue_addr, GEOCLUE_ADDRESS_KEY_REGION);
	locality = g_hash_table_lookup (geoclue_addr, GEOCLUE_ADDRESS_KEY_LOCALITY);
	area = g_hash_table_lookup (geoclue_addr, GEOCLUE_ADDRESS_KEY_AREA);
	postalcode = g_hash_table_lookup (geoclue_addr, GEOCLUE_ADDRESS_KEY_POSTALCODE);
	street = g_hash_table_lookup (geoclue_addr, GEOCLUE_ADDRESS_KEY_STREET);

	addr = location_address_new (NULL, street, area, locality, region, NULL, postalcode, NULL, country, NULL);
	return addr;
}


static LocationAccuracy *
_get_location_acc_from_geoclue_acc (GeoclueAccuracy *acc)
{
	if (!acc) return NULL;

	LocationAccuracy *accuracy = NULL;
	GeoclueAccuracyLevel level;
	double horiz_acc;
	double vert_acc;
	geoclue_accuracy_get_details (acc, &level, &horiz_acc, &vert_acc);

	if (level == GEOCLUE_ACCURACY_LEVEL_DETAILED)        accuracy = location_accuracy_new (LOCATION_ACCURACY_LEVEL_DETAILED, horiz_acc, vert_acc);
	else if (level == GEOCLUE_ACCURACY_LEVEL_STREET)     accuracy = location_accuracy_new (LOCATION_ACCURACY_LEVEL_STREET, horiz_acc, vert_acc);
	else if (level == GEOCLUE_ACCURACY_LEVEL_POSTALCODE) accuracy = location_accuracy_new (LOCATION_ACCURACY_LEVEL_POSTALCODE, horiz_acc, vert_acc);
	else if (level == GEOCLUE_ACCURACY_LEVEL_LOCALITY)   accuracy = location_accuracy_new (LOCATION_ACCURACY_LEVEL_LOCALITY, horiz_acc, vert_acc);
	else if (level == GEOCLUE_ACCURACY_LEVEL_REGION)     accuracy = location_accuracy_new (LOCATION_ACCURACY_LEVEL_REGION, horiz_acc, vert_acc);
	else if (level == GEOCLUE_ACCURACY_LEVEL_COUNTRY)    accuracy = location_accuracy_new (LOCATION_ACCURACY_LEVEL_COUNTRY, horiz_acc, vert_acc);
	else accuracy = location_accuracy_new (LOCATION_ACCURACY_LEVEL_NONE, 0, 0);

	return accuracy;
}

static void
address_cb (GeoclueReverseGeocode *revgeocode,
	GHashTable            *details,
	GeoclueAccuracy       *accuracy,
	GError                *error,
	gpointer               userdata)
{
	OsmHandleData* data = (OsmHandleData*)userdata;
	g_return_if_fail(data->addr_cb);

	LocationAddress *addr = NULL;
	LocationAccuracy *acc = NULL;
	LocationError err = LOCATION_ERROR_NONE;

	if (error) {
		MOD_LOGW ("Error getting reverse geocode: %s", error->message);
		err = LOCATION_ERROR_NOT_AVAILABLE;
	} else {
		if (accuracy) acc = _get_location_acc_from_geoclue_acc (accuracy);
		else acc = location_accuracy_new (LOCATION_ACCURACY_LEVEL_NONE, 0, 0);

		if (details) addr = _get_location_addr_from_geoclue_addr (details);
	}

	data->addr_cb (err, addr, acc, data->userdata);
	if (addr) location_address_free (addr);
	if (acc) location_accuracy_free (acc);
	g_free (data);
}

static void
_free_pos_list_item(gpointer data)
{
	g_return_if_fail(data);

	LocationPosition *position = (LocationPosition *)data;
	location_position_free(position);
}

static void
_free_acc_list_item(gpointer data)
{
	g_return_if_fail(data);

	LocationAccuracy *accuracy = (LocationAccuracy *)data;
	location_accuracy_free(accuracy);
}

static void
position_cb (GeoclueGeocode       *geocode,
	GeocluePositionFields fields,
	double                latitude,
	double                longitude,
	double                altitude,
	GeoclueAccuracy      *accuracy,
	GError               *error,
	gpointer              userdata)
{
	OsmHandleData* data = (OsmHandleData*)userdata;
	GList *pos_list = NULL;
	GList *acc_list = NULL;

	g_return_if_fail(data->pos_cb);

	LocationPosition *pos = NULL;
	LocationAccuracy *acc = NULL;
	LocationError err = LOCATION_ERROR_NONE;

	if (error) {
		MOD_LOGW ("Error getting geocode: %s", error->message);
		err = LOCATION_ERROR_NOT_AVAILABLE;
	} else {
		if (fields & GEOCLUE_POSITION_FIELDS_LATITUDE &&
			fields & GEOCLUE_POSITION_FIELDS_LONGITUDE) {
			if(fields & GEOCLUE_POSITION_FIELDS_ALTITUDE) pos = location_position_new (0, latitude, longitude, altitude, LOCATION_STATUS_3D_FIX);
			else pos = location_position_new (0, latitude, longitude, 0, LOCATION_STATUS_2D_FIX);

			pos_list = g_list_append(pos_list, pos);
		} else {
			err = LOCATION_ERROR_NOT_FOUND;
		}

		if (accuracy) {
			acc = _get_location_acc_from_geoclue_acc (accuracy);
			acc_list = g_list_append(acc_list, acc);
		}
	}
	data->pos_cb (err, pos_list, acc_list, data->userdata);

	if (pos_list) g_list_free_full(pos_list, _free_pos_list_item);
	if (acc_list) g_list_free_full(acc_list, _free_acc_list_item);
	g_free (data);

}


int
osm_get_geocode (OsmHandle *osm,
	const LocationAddress *addr,
	GList **pos_list,
	GList **acc_list)
{
	double lat, lon, alt;
	GeoclueAccuracy *geoclue_acc = NULL;
	GError *error = NULL;
	LocationPosition *pos = NULL;
	LocationAccuracy *acc = NULL;
	LocationError err = LOCATION_ERROR_NONE;

	GHashTable* geoclue_addr = _get_geoclue_addr_from_location_addr (addr);
	GeocluePositionFields fields = geoclue_geocode_address_to_position (osm->geocoder, geoclue_addr,
			                                      &lat, &lon, &alt,
			                                      &geoclue_acc, &error);
	g_hash_table_destroy (geoclue_addr);

	if (error) {
		MOD_LOGW ("Error getting geocode: %s", error->message);
		g_error_free (error);
		return LOCATION_ERROR_NOT_AVAILABLE;
	}

	if (fields & GEOCLUE_POSITION_FIELDS_LATITUDE &&
		fields & GEOCLUE_POSITION_FIELDS_LONGITUDE) {
		if(fields & GEOCLUE_POSITION_FIELDS_ALTITUDE) pos = location_position_new (0, lat, lon, alt, LOCATION_STATUS_3D_FIX);
			else pos = location_position_new (0, lat, lon, 0, LOCATION_STATUS_2D_FIX);

		*pos_list = g_list_append(*pos_list, pos);
	}
	else {
		err = LOCATION_ERROR_NOT_FOUND;
	}

	if (geoclue_acc) {
		acc = _get_location_acc_from_geoclue_acc (geoclue_acc);
		geoclue_accuracy_free (geoclue_acc);

		*acc_list = g_list_append(*acc_list, acc);
	}

	if (pos) location_position_free (pos);
	if (acc) location_accuracy_free (acc);

	return err;
}

int
osm_get_geocode_async (OsmHandle *osm,
	const LocationAddress * addr,
	LocationPositionCB callback,
	gpointer userdata)
{
	GHashTable* geoclue_addr = _get_geoclue_addr_from_location_addr (addr);
	OsmHandleData *data = g_new0 (OsmHandleData, 1);
	data->pos_cb = callback;
	data->userdata = userdata;
	geoclue_geocode_address_to_position_async (osm->geocoder, geoclue_addr, position_cb, data);
	g_hash_table_destroy (geoclue_addr);

	return LOCATION_ERROR_NONE;
}

int
osm_get_geocode_freetext(OsmHandle *osm,
	const const gchar* addr,
	GList **pos_list,
	GList **acc_list)
{
	double lat, lon, alt;
	GeoclueAccuracy *geoclue_acc = NULL;
	GError *error = NULL;
	LocationPosition *pos = NULL;
	LocationAccuracy *acc = NULL;
	LocationError err = LOCATION_ERROR_NONE;

	GeocluePositionFields fields = geoclue_geocode_freeform_address_to_position (osm->geocoder, addr,
			                                      &lat, &lon, &alt,
			                                      &geoclue_acc, &error);
	if (error) {
		MOD_LOGW ("Error getting geocode: %s", error->message);
		g_error_free (error);
		return LOCATION_ERROR_NOT_AVAILABLE;
	}

	if (fields & GEOCLUE_POSITION_FIELDS_LATITUDE &&
		fields & GEOCLUE_POSITION_FIELDS_LONGITUDE) {
		if(fields & GEOCLUE_POSITION_FIELDS_ALTITUDE) pos = location_position_new (0, lat, lon, alt, LOCATION_STATUS_3D_FIX);
		else pos = location_position_new (0, lat, lon, 0, LOCATION_STATUS_2D_FIX);
		*pos_list = g_list_append(*pos_list, pos);
	}
	else {
		err = LOCATION_ERROR_NOT_FOUND;
	}

	if (geoclue_acc) {
		acc = _get_location_acc_from_geoclue_acc (geoclue_acc);
		geoclue_accuracy_free (geoclue_acc);
		*acc_list = g_list_append(*acc_list, acc);
	}

	if (pos) location_position_free (pos);
	if (acc) location_accuracy_free (acc);
	return err;
}

int
osm_get_geocode_freetext_async (OsmHandle *osm,
	const gchar* addr,
	LocationPositionCB callback,
	gpointer userdata)
{
	OsmHandleData *data = g_new0 (OsmHandleData, 1);
	data->pos_cb = callback;
	data->userdata = userdata;
	geoclue_geocode_freeform_address_to_position_async (osm->geocoder, addr, position_cb, data);

	return LOCATION_ERROR_NONE;
}

int
osm_get_reverse_geocode(OsmHandle *osm,
	const LocationPosition *pos,
	LocationAddress **addr,
	LocationAccuracy **acc)
{
	GeoclueAccuracy *addr_acc = NULL;
	GError *error = NULL;
	GHashTable* geoclue_addr = NULL;

	double lat = pos->latitude;
	double lon = pos->longitude;
	GeoclueAccuracy *pos_acc = geoclue_accuracy_new (GEOCLUE_ACCURACY_LEVEL_DETAILED, 0.0, 0.0);

	gboolean success = geoclue_reverse_geocode_position_to_address (osm->rev_geocoder, lat, lon,
									pos_acc, &geoclue_addr,
									&addr_acc, &error);
	geoclue_accuracy_free (pos_acc);

	if (!success || error) {
		MOD_LOGW ("Error getting reverse geocode: %s", error->message);
		g_error_free (error);
		return LOCATION_ERROR_NOT_AVAILABLE;
	}

	if (addr_acc) {
		*acc = _get_location_acc_from_geoclue_acc (addr_acc);
		geoclue_accuracy_free (addr_acc);
	} else *acc = location_accuracy_new (LOCATION_ACCURACY_LEVEL_NONE, 0, 0);
	if (geoclue_addr) {
		*addr = _get_location_addr_from_geoclue_addr (geoclue_addr);
		g_hash_table_destroy (geoclue_addr);
	}

	return LOCATION_ERROR_NONE;
}

int
osm_get_reverse_geocode_async (OsmHandle *osm,
	const LocationPosition *pos,
	LocationAddressCB callback,
	gpointer userdata)
{
	double lat = pos->latitude;
	double lon = pos->longitude;
	GeoclueAccuracy *pos_acc = geoclue_accuracy_new (GEOCLUE_ACCURACY_LEVEL_DETAILED, 0.0, 0.0);;
	OsmHandleData *data = g_new0 (OsmHandleData, 1);
	data->addr_cb = callback;
	data->userdata = userdata;
	geoclue_reverse_geocode_position_to_address_async (osm->rev_geocoder,
						lat, lon, pos_acc,
						address_cb, data);
	geoclue_accuracy_free (pos_acc);

	return LOCATION_ERROR_NONE;
}




