/*
 * location-module
 *
 * Copyright (c) 2010-2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Youngae Kang <youngae.kang@samsung.com>, Minjune Kim <sena06.kim@samsung.com>
 *          Genie Kim <daejins.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib.h>
#include <stdlib.h>
#include <stdio.h>

#include <vconf.h>
#include <location.h>
#include <location-map-service.h>
#include <location-map-service-ext.h>
#include <geoclue/geoclue-poi.h>
#include <geoclue/geoclue-landmark.h>
#include "log.h"
#include "location-osm.h"
#include "location-osm-poi.h"

static int g_req_id = 0;

static void
_get_lang (gchar country_code[3], gchar lang_code[3])
{
	if (!country_code || !lang_code) return;
		gchar* langset = vconf_get_str(VCONFKEY_LANGSET);
		MOD_LOGW("getenv: %s", langset);

		if(langset == NULL){
			lang_code[0] = 'E';
			lang_code[1] = 'N';
			lang_code[2] = '\0';
			country_code[0] = 'U';
			country_code[1] = 'S';
			country_code[2] = '\0';
		}else{
			gchar* langset_upper = g_ascii_strup(langset, -1);
			lang_code[0] = langset_upper[0];
			lang_code[1] = langset_upper[1];
			lang_code[2] = '\0';
			country_code[0] = langset_upper[3];
			country_code[1] = langset_upper[4];
			country_code[2] = '\0';
			g_free(langset_upper);
		}
		MOD_LOGW("Language: %s, Country: %s", lang_code, country_code);
}

static void
_free_landmark(gpointer data)
{
	g_return_if_fail (data);

	LocationLandmark *landmark = (LocationLandmark *)data;
	location_landmark_free(landmark);
}


static void
_poi_callback(GeocluePoi *poi, int count, GPtrArray *landmark, GError *error, gpointer userdata)
{
	OsmHandleData *data = (OsmHandleData *)userdata;

	g_return_if_fail(data);
	g_return_if_fail(data->poi_cb);

	int req_id = data->req_id;
	int err = LOCATION_ERROR_NONE;
	GValueArray *vals = NULL;
	GList *landmark_list = NULL;
	LocationLandmark *land = NULL;
	GeoclueLandmark landmark_data = {0, };
	LocationBoundary *bounding_box = NULL;
	LocationPosition *left_top = NULL;
	LocationPosition *right_bottom = NULL;

	int idx = 0;
	if (!landmark || error) {
		MOD_LOGE("Landmark[%p] is NULL or Error occurs", landmark);
		err = LOCATION_ERROR_UNKNOWN;
	}
	else {
		MOD_LOGE("Count [%d]", count);
		for (; idx < count; idx++) {
			vals = (GValueArray *) g_ptr_array_index(landmark, idx);
			if (!vals) {
				break;
			}

			landmark_data.id = g_value_get_int (g_value_array_get_nth(vals, 0));
			landmark_data.rank = g_value_get_int (g_value_array_get_nth(vals, 1));
			landmark_data.lat = g_value_get_double (g_value_array_get_nth(vals, 2));
			landmark_data.lon = g_value_get_double (g_value_array_get_nth(vals, 3));
			landmark_data.boundary_left = g_value_get_double (g_value_array_get_nth(vals, 4));
			landmark_data.boundary_top = g_value_get_double (g_value_array_get_nth(vals, 5));
			landmark_data.boundary_right = g_value_get_double (g_value_array_get_nth(vals, 6));
			landmark_data.boundary_bottom = g_value_get_double (g_value_array_get_nth(vals, 7));
			landmark_data.name = g_value_dup_string (g_value_array_get_nth(vals, 8));
			landmark_data.icon = g_value_dup_string (g_value_array_get_nth(vals, 9));
			landmark_data.house = g_value_dup_string (g_value_array_get_nth(vals, 10));
			landmark_data.road = g_value_dup_string (g_value_array_get_nth(vals, 11));
			landmark_data.village = g_value_dup_string (g_value_array_get_nth(vals, 12));
			landmark_data.suburb = g_value_dup_string (g_value_array_get_nth(vals, 13));
			landmark_data.postcode = g_value_dup_string (g_value_array_get_nth(vals, 14));
			landmark_data.city = g_value_dup_string (g_value_array_get_nth(vals, 15));
			landmark_data.county = g_value_dup_string (g_value_array_get_nth(vals, 16));
			landmark_data.country = g_value_dup_string (g_value_array_get_nth(vals, 17));
			landmark_data.country_code = g_value_dup_string (g_value_array_get_nth(vals, 18));

			land = location_landmark_new();

			location_landmark_set_id (land, (guint)landmark_data.id);
			location_landmark_set_priority(land, (guint)landmark_data.rank);
			location_landmark_set_position (land, location_position_new (0, landmark_data.lat, landmark_data.lon, 0.0, LOCATION_STATUS_2D_FIX));

			location_landmark_set_name (land, landmark_data.name);
			location_landmark_set_address (land, location_address_new (NULL, landmark_data.house, landmark_data.village, landmark_data.city, landmark_data.county, NULL, landmark_data.postcode, NULL, landmark_data.country_code, NULL));
			location_landmark_set_property (land, "ICON", landmark_data.icon);

			if (!landmark_data.boundary_left || landmark_data.boundary_top || landmark_data.boundary_right || landmark_data.boundary_bottom) {
				left_top = location_position_new (0, landmark_data.boundary_left, landmark_data.boundary_top, 0.0, LOCATION_STATUS_2D_FIX);
				right_bottom = location_position_new (0, landmark_data.boundary_right, landmark_data.boundary_bottom, 0.0, LOCATION_STATUS_2D_FIX);
				bounding_box = location_boundary_new_for_rect (left_top, right_bottom);
				if (bounding_box) {
					location_landmark_set_bounding_box (land, bounding_box);
					location_boundary_free(bounding_box);
				}

				if (left_top) location_position_free (left_top);
				if (right_bottom) location_position_free (right_bottom);

				left_top = NULL;
				right_bottom = NULL;
				bounding_box = NULL;
			}


			landmark_list = g_list_append (landmark_list, land);
		}
	}

	MOD_LOGW ("Calling LocationPOICB");
	data->poi_cb (err, req_id, landmark_list, NULL, NULL, data->userdata);

	g_list_free_full (landmark_list, _free_landmark);
}

int
osm_search_poi_by_position(OsmHandle *osm,
			const LocationPOIFilter 	*filter,
			const LocationPosition 		*position,
			const LocationMapPref 	*svc_pref,
			const LocationPOIPreference 	*pref,
			LocationPOICB 			cb,
			const gpointer 			user_data,
			guint 				*req_id)
{
	int left = 0.0, top = 0.0, right = 0.0, bottom = 0.0;
	*req_id = ++g_req_id;

	OsmHandleData *data = g_new0 (OsmHandleData, 1);
	data->req_id = *req_id;
	data->poi_cb = cb;
	data->userdata = user_data;

	gchar *keyword = NULL, *lang = NULL, *country_code = NULL;
	int max_result = 0;

	keyword = location_poi_filter_get(filter, "KEYWORD");
	if (!keyword) {
		MOD_LOGW("OSM supports only keyword filter.");
		return LOCATION_ERROR_PARAMETER;

	}

	lang = location_map_pref_get_language (svc_pref);
	country_code = location_map_pref_get_country(svc_pref);
	max_result = location_poi_pref_get_max_result(pref);

	geoclue_poi_search_by_position_async (osm->poi, keyword, lang, country_code, max_result, left, top, right, bottom, _poi_callback, data);

	return LOCATION_ERROR_NONE;
}

int
osm_search_poi_by_area(OsmHandle *osm,
			const LocationPOIFilter 	*filter,
			const LocationBoundary 		*boundary,
			const LocationMapPref 	*svc_pref,
			const LocationPOIPreference 	*pref,
			LocationPOICB 			cb,
			const gpointer 			user_data,
			guint 				*req_id)
{
	*req_id = ++g_req_id;

	OsmHandleData *data = g_new0 (OsmHandleData, 1);

	gchar sys_country[3] = {0, }, sys_lang[3] = {0, };
	gchar *keyword = NULL, *lang = NULL, *country_code = NULL;
	int max_result = 0;
	gdouble left = 0.0, top = 0.0, right = 0.0, bottom = 0.0;

	data->req_id = *req_id;
	data->poi_cb = cb;
	data->userdata = user_data;

	_get_lang (sys_country, sys_lang);
	keyword = location_poi_filter_get(filter, "KEYWORD");
	if (!keyword) {
		MOD_LOGW("OSM supports only keyword filter.");
		return LOCATION_ERROR_PARAMETER;

	}

	lang = location_map_pref_get_language (svc_pref);
//	if (!lang) lang = sys_lang;

	country_code = location_map_pref_get_country(svc_pref);
//	if (!country_code) country_code = sys_country;

	max_result = location_poi_pref_get_max_result(pref);

	LocationPosition *left_top = boundary->rect.left_top;
	LocationPosition *right_bottom = boundary->rect.right_bottom;

	left = left_top->latitude;
	top = left_top->longitude;
	right = right_bottom->latitude;
	bottom = right_bottom->longitude;

	geoclue_poi_search_by_position_async (osm->poi, keyword, lang, country_code, max_result, left, top, right, bottom, _poi_callback, data);

	return LOCATION_ERROR_NONE;
}
