/*
 * location-module
 *
 * Copyright (c) 2010-2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Youngae Kang <youngae.kang@samsung.com>, Minjune Kim <sena06.kim@samsung.com>
 *          Genie Kim <daejins.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib.h>
#include <json-glib.h>
#include <stdlib.h>

#include <location.h>
#include <location-map-service.h>
#include <location-map-service-ext.h>
#include "log.h"
#include "osm-route-yours.h"
#include "location-osm-route.h"
#include "osm-http.h"

static char *
_create_route_url (const LocationPosition *origin, const LocationPosition *destination, const GList *waypoint, const LocationMapPref *svc_pref, const LocationRoutePreference *pref)
{
	gdouble flon = 0.0, flat = 0.0, tlon = 0.0, tlat = 0.0;
	GList *ways = NULL;
	char *mode = "motorcar";


	flon = origin->longitude;
	flat = origin->latitude;
	tlon = destination->longitude;
	tlat = destination->latitude;

#if 0
	if (waypoint)
#endif

	return yournavi_create_url (flat, flon, tlat, tlon, ways, 1, mode, 1);
	//yournavi_create_url (gdouble flon, gdouble flat, gdouble tlon, gdouble tlat, const GList *ways, int type, char *mode, int instruction);

}

static void
_position_foreach_cb (gpointer data, gpointer user_data)
{
	if (!data) return ;

	osm_data_t *ndata = (osm_data_t *)user_data;

	LocationPosition *pos = (LocationPosition*) data;
	LocationPosition *duplicated = location_position_copy (pos);

	ndata->waypoint = g_list_append(ndata->waypoint, duplicated);

}

#if 0
static void
_position_destory_cb (gpointer data)
{
	if (!data) return;

	LocationPosition *pos = (LocationPosition *)data;

	location_position_free (pos);
}
#endif

static void
_get_position_from_coordinates (JsonArray *_array, guint index, JsonNode *element, gpointer user_data)
{
	GList **pos_list = (GList **)user_data;
	gdouble lat = 0.0, lon = 0.0;

	JsonArray *array = json_node_get_array(element);
	if (array) {
		lon = json_array_get_double_element (array, 0);
		lat = json_array_get_double_element (array, 1);
		*pos_list = g_list_append(*pos_list, location_position_new(0, lat, lon, 0.0, LOCATION_STATUS_2D_FIX));
	}

}

int
convert_data_to_location_service (const char *result, int result_len, LocationPosition *origin, LocationPosition *dest, GList **route_list)
{
	gboolean ret = FALSE;
	GError *err = NULL;
	JsonParser* parser = json_parser_new ();
	JsonNode* node = NULL;
	JsonObject *obj = NULL;
	JsonObject *properties = NULL;
	JsonArray *positions = NULL;
//	gchar *instructions = NULL;

	GList *pos_list = NULL;
	LocationRoute *route = NULL;
	gchar *str_distance = NULL, *str_duration = NULL;
	gdouble distance = 0.0;
	glong duration = 0;

	if (!result) return -1;

	ret = json_parser_load_from_data(parser, result, result_len, &err);
	if (ret == FALSE) {
		MOD_LOGW("Fail to parser data");
		return -1;
	}

	node = json_parser_get_root(parser);
	if (!node) {
		MOD_LOGW("Fail to parser data");
		return -1;
	}

	obj = json_node_get_object(node);
	if (!obj) {
		MOD_LOGW("Fail to parser data");
		return -1;
	}

	positions = json_object_get_array_member(obj, "coordinates");
	if (positions) {
		json_array_foreach_element (positions, _get_position_from_coordinates, &pos_list);
		MOD_LOGW("foreach");
	}

	properties = json_object_get_object_member(obj, "properties");
	if (properties) {
		str_distance  = (gchar *)json_object_get_string_member(properties, "distance");
		MOD_LOGW("distance [%s]", str_distance);
	//	instructions = json_object_get_string_member (properties, "description");
		str_duration = (gchar *)json_object_get_string_member (properties, "traveltime");
		MOD_LOGW("time [%s]", str_duration);

		if (str_distance) {
			distance = atof (str_distance);
			MOD_LOGW("converted distance  [%f]", distance);
			g_free(str_distance);
		}

		if (str_duration) {
			duration = atol (str_duration);
			MOD_LOGW("converted duration  [%ld]", duration);
			g_free( str_duration);
		}
	}
	else {
		MOD_LOGW("property NULL");
	}

	LocationRouteStep *step = location_route_step_new ();

	location_route_step_set_start_point (step, location_position_copy (origin));
	location_route_step_set_end_point (step, location_position_copy (dest));


	if (pos_list) {
		location_route_step_set_geometry (step, pos_list);
	}
	GList *step_list = NULL;
	step_list = g_list_append (step_list, step);


	// Route == RouteSegment
	LocationRouteSegment *segment = location_route_segment_new();
	location_route_segment_set_start_point (segment, location_position_copy (origin));
	location_route_segment_set_end_point (segment, location_position_copy (dest));
	location_route_segment_set_distance (segment, distance);
	location_route_segment_set_duration (segment, duration);
	location_route_segment_set_route_step (segment, step_list);

	GList *seg_list = NULL;
	seg_list = g_list_append (seg_list, segment);


	route = location_route_new ();
	location_route_set_origin(route, origin);
	location_route_set_destination(route, dest);
	location_route_set_total_distance(route, distance);
	location_route_set_distance_unit(route, "M");
	location_route_set_total_duration(route, duration);
	location_route_set_route_segment(route, seg_list);

	*route_list = g_list_append(*route_list, route);

	return 0;
}

int
osm_request_route (const LocationPosition *origin,
			const LocationPosition *destination,
			GList *waypoint,
			const LocationMapPref *svc_pref,
			const LocationRoutePreference *pref,
			LocationRouteCB cb,
			const gpointer user_data,
			guint *req_id)
{
	osm_data_t *ndata = g_new0(osm_data_t, 1);

	ndata->route_cb = cb;
	ndata->user_data = (gpointer)user_data;
	ndata->origin = location_position_copy(origin);
	ndata->dest = location_position_copy(destination);
	if (waypoint) g_list_foreach(waypoint, _position_foreach_cb, ndata);

	char *url = _create_route_url (origin, destination, waypoint, svc_pref, pref);
	if (!url) {
		g_free(ndata);
		return LOCATION_ERROR_UNKNOWN;
	}

	int ret = curl_request_route_async(url, ndata, req_id);
	if (ret != 0) {
		g_free(ndata);
		g_free(url);
		return LOCATION_ERROR_UNKNOWN;
	}

	g_free(ndata);
	g_free(url);
	return LOCATION_ERROR_NONE;
}

int
osm_cancel_route_request (guint req_id)
{
	return curl_cancel_route_request (req_id);
}
