/*
 * location-module
 *
 * Copyright (c) 2010-2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Youngae Kang <youngae.kang@samsung.com>, Minjune Kim <sena06.kim@samsung.com>
 *          Genie Kim <daejins.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <curl/curl.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include <pthread.h>
#include <unistd.h>
#include "log.h"
#include "location-osm-route.h"
#include "osm-http.h"

//http://www.yournavigation.org/api/1.0/gosmore.php?format=geojson&flat=52.215676&flon=5.963946&tlat=52.2573&tlon=6.1799&v=motorcar&fast=1&layer=mapnik


#define YOUR_NAVIGATION_SERVER	"http://www.yournavigation.org/api/1.0/gosmore.php?"
#define YOUR_NAVIGATION_FORMAT	"format="

#define KML_FORMAT "kml"
#define GEOJSON_FORMAT "geojson"

static CURLM *g_multi_handle;
static pthread_t g_http_thread;
static GHashTable *g_req_table;

pthread_mutex_t curl_mutex = PTHREAD_MUTEX_INITIALIZER;

typedef int (*curl_async_cb) (int req_id, int error, const char *result, int result_len, void *user_data);

typedef struct {
	 guint req_id;
	CURL *ctx;
	curl_async_cb cb;
	void *user_data;

	int result_len;
	char *result;
} curl_item_t;

CURL*
curl_create_session (guint *req_id)
{
	static guint session_id = 0;

	CURL_MUTEX_LOCK(&curl_mutex);
	if (!g_multi_handle)
		g_multi_handle = curl_multi_init();
	CURL_MUTEX_UNLOCK(&curl_mutex);

	CURL* ctx = curl_easy_init();
	if (!ctx) {
		MOD_LOGW("Unable to initialize cURL interface\n");
		return NULL;
	}

	*req_id = ++session_id;

	MOD_LOGW ("req_id [%d]", *req_id);

	return ctx;
}

int
curl_destroy_session (CURL* ctx)
{
	if (!ctx) return -1;

	curl_easy_cleanup(ctx);

	return 0;
}

static gboolean
_compare_ctx (gpointer key, gpointer value, gpointer userdata)
{
	if (!key || !value || !userdata) return FALSE;

	curl_item_t * item = (curl_item_t*) value;
	CURL *ctx = (CURL*) userdata;

	if (item->ctx == ctx) {
		return TRUE;
	}

	return FALSE;
}

static int
_send_resp (curl_item_t *item)
{
	MOD_LOGW ("Send response [%p]", item);
	if (!item) return -1;
	MOD_LOGW("Result [%s]", item->result);

	if (item->cb)
		item->cb (item->req_id, 0, item->result, item->result_len, item->user_data);

	if (item->user_data) {
		g_free(item->user_data);
		item->user_data = NULL;
	}

	if (item->result) {
		g_free(item->result);
		item->result = NULL;
	}

	return 0;
}

static void *
_create_http_thread(void *data)
{
	MOD_LOGW ("Create Thread");
	CURLMsg *msg = NULL;
	int msgs_left = 0;
	int still_running = 1;
	int rc = 0;
	int ret = 0;
	struct timeval timeout;

	timeout.tv_sec = 10;
	timeout.tv_usec = 0;

	while(still_running) {
		fd_set fdread;
		fd_set fdwrite;
		fd_set fdexecp;
		int maxfd = -1;

		FD_ZERO(&fdread);
		FD_ZERO(&fdwrite);
		FD_ZERO(&fdexecp);

		if (g_req_table)
			curl_multi_fdset (g_multi_handle, &fdread, &fdwrite, &fdexecp, &maxfd);
		else {
			MOD_LOGW ("Hash table is NULL");
			break;
		}

		rc = select(maxfd+1, &fdread, &fdwrite, &fdexecp, &timeout);
		MOD_LOGW ("run loop [%d]", still_running);
		switch (rc) {
			case -1:
				break;
			case 0:
			default:
				/* Timeout or readable / writable sockets */
				curl_multi_perform(g_multi_handle, &still_running);
				if (ret != CURLM_OK) {
					MOD_LOGW ("curl_multi_perform error [%d]", ret);
					break;
				}

				msg = curl_multi_info_read(g_multi_handle, &msgs_left);
				if (msg && msg->msg == CURLMSG_DONE) {
					curl_item_t *item = g_hash_table_find (g_req_table, _compare_ctx, msg->easy_handle);
					if (item) {
						MOD_LOGW ("Found Item ID[%d]", item->req_id);
						int ret = _send_resp(item);
						if (ret) {
							MOD_LOGW ("Fail to send response [%d]", ret);
							g_http_thread = 0;
							return NULL;
						}

						g_hash_table_remove (g_req_table, (gpointer) &item->req_id);

						curl_destroy_session(item->ctx);
						if (item) {
							g_free(item);
							item = NULL;
						}

					}
				}
				else {
					sleep(1);
				}
				break;
		}
	}

	CURL_MUTEX_LOCK (&curl_mutex);
	if (g_multi_handle) {
		curl_multi_cleanup(g_multi_handle);
		g_multi_handle = NULL;
	}
	CURL_MUTEX_UNLOCK (&curl_mutex);

	MOD_LOGW("End Thread");
	g_http_thread = 0;

	return NULL;
}

static size_t
_osm_route_cb (char *ptr, size_t size, size_t nmemb, void *userdata)
{
	curl_item_t *item = (curl_item_t*)userdata;

	MOD_LOGW("size[%d], nmemb [%d]", size, nmemb);
	int reallo_len = size * nmemb;

	item->result = realloc (item->result, item->result_len + reallo_len + 1);
	if (item->result == NULL) {
		/* out of memory! */
		MOD_LOGW("not enough memory (realloc returned NULL)\n");
		return 0;
	}

	memcpy(&(item->result[item->result_len]), ptr, reallo_len);
	item->result_len += reallo_len;
	item->result[item->result_len] = 0;

	return reallo_len;
}

static int
_route_cb (int req_id, int error, const char *result, int result_len, void *user_data)
{
	GList *route = NULL;
	char *error_code = NULL;
	char *error_msg = NULL;

	osm_data_t *ndata = (osm_data_t *) user_data;
	if (result && result_len > 0) {
		// parsing
	}

	convert_data_to_location_service (result, result_len, ndata->origin, ndata->dest, &route);


	ndata->route_cb(error, req_id, route, error_code, error_msg, ndata->user_data);

	return 0;
}

int
curl_request_route_async(const char *url, osm_data_t *data, guint *req_id)
{
	curl_item_t *item = g_new0(curl_item_t, 1);
	if (!item) return -999;

	int stillrunning = 0;


	CURL *ctx = curl_create_session (req_id);
	if (!ctx) {
		*req_id = -1;
		return -1;
	}


	CURLMcode err = 0;
	int ret = 0;

	item->req_id = *req_id;
	item->user_data = data;
	item->ctx = ctx;
	item->cb = _route_cb;

	curl_easy_setopt (ctx, CURLOPT_URL, url);
//	curl_easy_setopt (ctx, CURLOPT_NOPROGRESS , 0L);
//	curl_easy_setopt( ctx , CURLOPT_WRITEHEADER , stderr);
	curl_easy_setopt( ctx, CURLOPT_WRITEFUNCTION, _osm_route_cb);
	curl_easy_setopt( ctx , CURLOPT_WRITEDATA, item);

	CURL_MUTEX_LOCK (&curl_mutex);
	curl_multi_add_handle(g_multi_handle, ctx);

	if (!g_req_table) {
		g_req_table = g_hash_table_new (g_int_hash, g_int_equal);
	}

	g_hash_table_insert(g_req_table, req_id, item);

	MOD_LOGW ("ReqID: %u", *req_id);

	err = curl_multi_perform(g_multi_handle, &stillrunning);
	CURL_MUTEX_UNLOCK (&curl_mutex);
	if (CURLM_OK != err) {
		MOD_LOGW ("curl_multi_perform error[%d]", err);
		if (item) {
			g_free(item);
			item = NULL;
		}

		curl_destroy_session(ctx);
		return -1;
	}

	if (!g_http_thread) {	//TODO
		ret = pthread_create(&g_http_thread, NULL, _create_http_thread, NULL);

		if (ret) {
			// return error
			MOD_LOGW ("Fail to create thread [%d", ret);
			if (item) {
				g_free(item);
				item = NULL;
			}

			curl_destroy_session(ctx);
			return -1;
		}
		MOD_LOGW ("End loop");
		pthread_detach(g_http_thread);
	}
	else {
		MOD_LOGW ("Thread is alive");
	}

	return 0;
}

int
curl_cancel_route_request (guint req_id)
{
	curl_item_t* item = NULL;

	if (!g_req_table) {
		return LOCATION_ERROR_NOT_FOUND;
	}

	item = (curl_item_t *)g_hash_table_lookup (g_req_table, (void* )&req_id);
	if (!item) {
		return LOCATION_ERROR_NOT_FOUND;
	}


	if (g_multi_handle) {
		curl_multi_remove_handle (g_multi_handle, item->ctx);
	}

	if( item->user_data) {
		g_free (item->user_data);
		item->user_data = NULL;
	}

	if (item->result) {
		g_free (item->result);
		item->result = NULL;
	}
	if (item) {
		g_free (item);
		item = NULL;
	}

	gboolean ret = g_hash_table_remove(g_req_table, (void *)&req_id);
	if (!ret) {
		return LOCATION_ERROR_NOT_FOUND;
	}
	if (g_hash_table_size(g_req_table) == 0) {
		MOD_LOGW ("Hash table is NULL");
		g_hash_table_destroy(g_req_table);
		g_req_table = NULL;
	}

	return LOCATION_ERROR_NONE;
}
